import typer
from decorator import decorator

_CONTEXT: typer.Context = None


@decorator
def set_context(func, *args, **kwargs):
    global _CONTEXT
    _CONTEXT = args[0]
    return func(*args, **kwargs)


def get_context():
    return _CONTEXT



from typing import Annotated

from click import ClickException
from pydantic import Field, validate_arguments

from pekora.context import get_context


class PekoraProblem(Exception):
    """
    Base exception for Pekora.
    """


class Otsupeko(PekoraProblem, ClickException):
    """
    An exception that signals Pekora to perform a nonzero exit with an error message.

    To exit code 0 or without a message, use :class:`typer.Exit()` instead.

    Parameters
    ----------
    message : str
        The error message.
    code : int, default: 1
        The code to exit with.
    """

    @validate_arguments
    def __init__(self, message: str, code: Annotated[int, Field(ge=1, le=255)] = 1):
        ClickException.__init__(self, message.strip())
        setattr(self, "ctx", get_context())
        self.exit_code = code



from pathlib import Path
from typing import ClassVar, Self

from pydantic import BaseSettings

from pekora import utils


class PekoraPreferences(BaseSettings):
    class Config:
        extra = "forbid"

    file: ClassVar[Path] = utils.pekora_home() / "config.json"

    debug: bool = False

    @classmethod
    def load(cls) -> Self:
        if not cls.file.exists():
            cls.file.write_text(cls().json())

        return cls.parse_file(cls.file)

    def save(self):
        self.file.write_text(self.json())

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()



from __future__ import annotations

import re
import sys
from datetime import datetime
from functools import partial

import alianator
import inflect as ifl
import pyperclip
import typer
from InquirerPy.base import Choice
from rich import print, print_json
from rich.panel import Panel
from rich.table import Table

from pekora import callbacks, prompts, utils
from pekora.context import set_context
from pekora.exceptions import *
from pekora.models import *
from pekora.peko import Pekora

app = Pekora()
inflect = ifl.engine()


@app.command(name="calc", no_args_is_help=True)
@set_context
def calculate(
    _: typer.Context = None,
    expression: str = typer.Argument(
        ..., help="The expression to evaluate.", show_default=False
    ),
    raw: bool = typer.Option(
        None,
        "--raw",
        "-r",
        help="Don't use pretty formatting in the output. Ideal for piping to [bold cyan]pekora read[/] or "
        "other commands.",
    ),
    copy: bool = typer.Option(
        None,
        "--copy",
        "-c",
        help="Copy the result to the clipboard.",
    ),
):
    """
    Evaluate an expression.
    """

    def evaluate(expr: str) -> PekoraPermissions:
        pattern = re.compile("|".join((str(p) for p in PekoraPattern.all())))
        return eval(pattern.sub(lambda x: str(utils.ninjin(x.group())), expr))

    # Split the expression on its comparators.
    parts = list(filter(None, re.split(rf"({PekoraPattern.COMPARATOR})", expression)))

    # Balanced comparators.
    if len(parts) % 2:
        if len(parts) > 3 and set(parts).intersection({"==", "!="}):
            raise Otsupeko(
                "An equality comparator may not be used in the same expression as other comparators "
                "(including other equality comparators).",
            )

        result = eval(
            "".join(
                [
                    part if index % 2 else str(evaluate(part))
                    for index, part in enumerate(parts)
                ]
            )
        )

    # Imbalanced comparators.
    else:
        raise Otsupeko("Comparators must have an expression on both sides.")

    print(
        result
        if raw
        else Panel(
            f"[cyan]{result}[/]", title="Result", title_align="left", style="green"
        )
    )

    if copy:
        pyperclip.copy(result)


@app.command(name="read", no_args_is_help=True)
@set_context
def read(
    _: typer.Context = None,
    permission: str = typer.Argument(
        ...,
        help="A permission flag, integer value, or Pekora permission group.",
        show_default=False,
        allow_dash=True,
    ),
    include: list[PekoraProperties.Type] = typer.Option(
        None,
        "--include",
        "--with",
        "-i",
        help="Explicitly include a data category, excluding all others not passed with -i.",
        show_default=False,
        rich_help_panel="Filter and Format",
    ),
    exclude: list[PekoraProperties.Type] = typer.Option(
        None,
        "--exclude",
        "--without",
        "-e",
        "-x",
        help="Explicitly exclude a data category. Supersedes -i.",
        show_default=False,
        rich_help_panel="Filter and Format",
    ),
    as_json: bool = typer.Option(
        None,
        "--json",
        help="Output the result as JSON.",
        rich_help_panel="Filter and Format",
    ),
    copy: bool = typer.Option(
        None,
        "--copy",
        "-c",
        help="Copy a JSON representation of the result to the clipboard (does not require --json).",
    ),
):
    """
    Read a permission.
    """
    if permission == "-":
        permission = sys.stdin.read()

    if not re.match(
        f"({'|'.join(map(str, PekoraPattern.permissions()))})$",
        permission,
    ):
        raise Otsupeko(f"Invalid permission value: {permission}")

    include = (set(include) or set(PekoraProperties.Type)) - set(exclude)

    if not include:
        raise Otsupeko("You must include at least one data category.")

    permset = PekoraPack.from_permissions(eval(utils.ninjin(permission)))

    json = permset.json(
        include={
            "permissions": {"__all__": {category.value for category in include}},
            "derived_from": True,
        }
    )

    if as_json:
        print_json(json)
    else:
        table_caption = f"Derived from: {permset.derived_from}"
        table = Table(
            caption=table_caption,
            min_width=len(table_caption),
            border_style=utils.pekora_blue().hex,
        )

        table.add_column("Flag", style="cyan")
        table.add_column("Name", style="yellow")
        table.add_column("Value", style="green")

        for perm in permset.permissions:
            table.add_row(*perm)

        for col in table.columns.copy():
            if not any(col.header.casefold() == category.value for category in include):
                table.columns.remove(col)

        print(table)

    if copy:
        pyperclip.copy(json)


# noinspection PyUnusedLocal
@app.command(name="make")
@set_context
def make(
    ctx: typer.Context = None,
    start: str = typer.Option(
        0,
        "--from",
        help="A permission flag, integer value, or Pekora permission group representing the permissions to start with.",
        show_default=False,
    ),
):
    """
    Interactively create a permission.
    """
    if not re.match(
        f"({'|'.join(map(str, PekoraPattern.permissions()))})$",
        start,
    ):
        raise Otsupeko(f"Invalid permission value: {start}")

    permissions = eval(utils.ninjin(start))

    choices = []
    for flag, name in alianator.resolutions(escape_mentions=False).items():
        if not any(
            PekoraPermissions.from_flags(c.value) == PekoraPermissions.from_flags(flag)
            for c in choices
        ):
            choices.append(
                Choice(value=flag, name=name, enabled=getattr(permissions, flag))
            )
    try:
        permissions += prompts.fuzzy(
            message="Choose some permissions. Type to search.",
            choices=choices,
            multiselect=True,
            border=True,
            transformer=lambda v: inflect.no("permission", len(v)),
            filter=lambda v: PekoraPermissions.from_flags(*v),
        ).execute()

        print(
            Panel(
                f"[cyan]{permissions}[/]",
                title="Result",
                title_align="left",
                style="green",
            )
        )

        prompts.select(
            message="What would you like to do with the result?",
            choices=[
                Choice(
                    value=partial(pyperclip.copy, str(permissions)),
                    name="Copy to clipboard",
                ),
                Choice(
                    value=partial(
                        read,
                        permission=str(permissions),
                        include=set(),
                        exclude=set(),
                        as_json=False,
                    ),
                    name="Read",
                ),
                Choice(
                    value=partial(make, ctx=ctx, start=str(permissions)),
                    name="Restart using this result as the starting value",
                ),
                Choice(value=lambda: ..., name="Nothing"),
            ],
        ).execute()()
    except KeyboardInterrupt:
        typer.Exit()


# noinspection PyUnusedLocal
@app.callback(epilog=f"Pekora © {datetime.now().year} celsius narhwal.")
def konpeko(
    docs: bool = typer.Option(
        None,
        "--docs",
        is_eager=True,
        help="View Pekora's documentation.",
        callback=callbacks.docs,
        rich_help_panel="About Pekora",
    ),
    version: bool = typer.Option(
        None,
        "--version",
        is_eager=True,
        help="View Pekora's version.",
        callback=callbacks.version,
        rich_help_panel="About Pekora",
    ),
    license: bool = typer.Option(
        None,
        "--license",
        is_eager=True,
        help="View Pekora's license.",
        callback=callbacks.license,
        rich_help_panel="About Pekora",
    ),
    repo: bool = typer.Option(
        None,
        "--repo",
        is_eager=True,
        help="Visit Pekora on GitHub.",
        callback=callbacks.repo,
        rich_help_panel="About Pekora",
    ),
    debug: bool = typer.Option(
        None,
        "--debug",
        is_eager=True,
        hidden=True,
        callback=callbacks.debug,
    ),
):
    """
    Pekora is a calculator for Discord permission values.
    """


if __name__ == "__main__":
    app()



from importlib import metadata
from pathlib import Path

import typer
from decorator import decorator
from rich import print
from rich.markdown import Markdown
from rich.panel import Panel

from pekora import utils
from pekora.prefs import PekoraPreferences


@decorator
def callback(func, execute: bool):
    if execute:
        func(execute)
        raise typer.Exit()


@callback
def docs(_):
    typer.launch("https://pekora.celsiusnarhwal.dev")


@callback
def version(_):
    print(
        Panel(
            f"Pekora [cyan]{metadata.version('pekora')}[/]",
            title="Version",
            title_align="left",
            border_style=utils.pekora_blue().hex,
        )
    )


@callback
def license(_):
    license_file = (
        next(p for p in Path(__file__).parents if (p / "LICENSE.md").exists())
        / "LICENSE.md"
    )

    print(
        Panel(
            Markdown(license_file.read_text()),
            title="License",
            border_style=utils.pekora_blue().hex,
        )
    )


@callback
def repo(_):
    typer.launch(str(utils.pekora_repo()))


@callback
def debug(_):
    with PekoraPreferences.load() as settings:
        settings.debug = not settings.debug
        status = "[green]enabled[/]" if settings.debug else "[red]disabled[/]"

    print(
        Panel(
            f"Debug mode {status}.",
            title="Debug Mode",
            title_align="left",
            border_style="violet",
        )
    )



from __future__ import annotations

from pathlib import Path
from typing import Callable

import typer
from colour import Color
from yarl import URL

from pekora.exceptions import *
from pekora.models import *


@validate_arguments
def ninjin(term: str) -> str:
    """
    Convert a Pekora expression term to a string that eval() can turn into a :class:`PekoraPermissions` object.

    Pekora expression terms are:
    - Integers
    - Discord permission flags (e.g. `read_messages`)
    - Pekora permission groups (e.g. `pekora.general`)

    Parameters
    ----------
    term : str
        A valid Pekora expression term.

    Returns
    -------
    str
        A string that eval() can turn into a :class:`PekoraPermissions` object.

    """
    # Handle integers.
    if PekoraPattern.INTEGER.regex.match(term):
        return f"PekoraPermissions({term})"

    # Handle Discord permission flags.
    if term in PekoraPermissions.VALID_FLAGS:
        return f"PekoraPermissions.from_flags('{term}')"

    # Handle Pekora permission groups.
    if match := PekoraPattern.GROUP.regex.match(term):
        group = match.group("group")
        try:
            if isinstance(method := getattr(PekoraPermissions, group), Callable):
                if isinstance(method(), PekoraPermissions):
                    return f"PekoraPermissions.{group}()"
                else:
                    raise TypeError
            else:
                raise AttributeError
        except (AttributeError, TypeError):
            raise Otsupeko(f"{term} is not a valid Pekora permission group.")

    # Handle unsupported operators.
    if PekoraPattern.UNSUPPORTED.regex.match(term):
        raise Otsupeko(f"Unsupported operator: {term}")

    # Handle other invalid input.
    raise Otsupeko(f"Invalid permission value: {term}")


def pekora_home() -> Path:
    home = Path(typer.get_app_dir("Pekora"))
    home.mkdir(exist_ok=True)
    return home


def pekora_repo() -> URL:
    return URL("https://github.com/celsiusnarhwal/pekora")


def pekora_blue() -> Color:
    return Color("#b0bfe9")


def debug_epilog() -> str:
    return "[dim]Debug mode is [green]on[/]. Turn it off with [bold cyan]pekora --debug[/]."



from functools import wraps

from InquirerPy import inquirer


def prompt_decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if kwargs.get("multiselect"):
            keybindings = {
                "toggle": [{"key": "c-s"}],
                "toggle-all-true": [{"key": "c-a"}],
                "toggle-all-false": [{"key": "c-z"}],
            }

            long_instruction = (
                "↑/↓: Move up and down\n"
                "Control + S: Toggle selection\n"
                "Control + A: Toggle all on\n"
                "Control + Z: Toggle all off\n"
                "Enter: Confirm\n"
            )

        else:
            keybindings = {
                "answer": [{"key": "enter"}, {"key": "c-z"}],
            }

            long_instruction = ""

        defaults = {
            "keybindings": keybindings,
            "long_instruction": long_instruction,
            "qmark": "•",
            "amark": "✓",
            "mandatory_message": "You can't skip this.",
            "raise_keyboard_interrupt": True,
        }

        for key, value in defaults.items():
            kwargs.setdefault(key, value)

        return func(*args, **kwargs)

    return wrapper


checkbox = prompt_decorator(inquirer.checkbox)
confirm = prompt_decorator(inquirer.confirm)
expand = prompt_decorator(inquirer.expand)
filepath = prompt_decorator(inquirer.filepath)
fuzzy = prompt_decorator(inquirer.fuzzy)
text = prompt_decorator(inquirer.text)
select = prompt_decorator(inquirer.select)
number = prompt_decorator(inquirer.number)
rawlist = prompt_decorator(inquirer.rawlist)
secret = prompt_decorator(inquirer.secret)



from __future__ import annotations

import re
from enum import StrEnum, auto
from typing import Self

import alianator
import discord
from pydantic import BaseModel, validate_arguments

__all__ = [
    "PekoraPermissions",
    "PekoraPattern",
    "PekoraProperties",
    "PekoraPack",
]


class PekoraPermissions(discord.Permissions):
    @property
    def flags(self) -> list[str]:
        return [perm for perm, granted in self if granted]

    @classmethod
    def from_flags(cls, *flags: str) -> Self:
        return cls(**{flag: True for flag in flags})

    def __str__(self):
        return str(self.value)

    def __iter__(self):
        cls = type(self)
        return cls.__base__.__iter__(cls.__base__(self.value))


class PekoraPattern(StrEnum):
    # Patterns must be ordered by match precedence.
    GROUP = r"pekora\.(?P<group>\w+)"
    INTEGER = r"-?\d+"
    FLAG = r"\w+"
    COMPARATOR = r"==|!=|[<>]=?"
    UNSUPPORTED = r"[*/%@=]"

    @classmethod
    def all(cls) -> tuple[Self]:
        return tuple(iter(cls))

    @classmethod
    def permissions(cls) -> tuple[Self]:
        return cls.GROUP, cls.INTEGER, cls.FLAG

    @property
    def regex(self) -> re.Pattern:
        return re.compile(self.value)

    def __str__(self):
        return self.value


class PekoraProperties(BaseModel):
    class Type(StrEnum):
        FLAG = auto()
        NAME = auto()
        VALUE = auto()

    flag: str
    name: str
    value: str

    def __iter__(self):
        return iter((self.flag, self.name, self.value))


class PekoraPack(BaseModel):
    derived_from: int
    permissions: list[PekoraProperties]

    @classmethod
    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def from_permissions(cls, permissions: PekoraPermissions):
        return cls(
            derived_from=permissions.value,
            permissions=[
                PekoraProperties(
                    flag=flag,
                    name=alianator.resolve(flag, escape_mentions=False)[0],
                    value=str(PekoraPermissions.from_flags(flag).value),
                )
                for flag in permissions.flags
            ],
        )



from __future__ import annotations

import typer
from rich import print
from rich.panel import Panel

from pekora import utils
from pekora.prefs import PekoraPreferences


class Pekora(typer.Typer):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("no_args_is_help", True)
        kwargs.setdefault("rich_markup_mode", "rich")
        super().__init__(*args, **kwargs)

    def callback(self, *args, **kwargs):
        with PekoraPreferences.load() as settings:
            if settings.debug:
                kwargs["epilog"] = (
                    kwargs.get("epilog", "") + "\n\n" + utils.debug_epilog()
                ).strip()

        return super().callback(*args, **kwargs)

    def command(self, *args, **kwargs):
        with PekoraPreferences.load() as settings:
            if settings.debug:
                kwargs["epilog"] = (
                    kwargs.get("epilog", "") + "\n\n" + utils.debug_epilog()
                ).strip()

        return super().command(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        try:
            super().__call__(*args, **kwargs)
        except Exception as e:  # Yes, we really do want to catch *everything*.
            if e is typer.Exit:  # "except typer.Exit: pass" doesn't work, apparently.
                return

            with PekoraPreferences.load() as settings:
                if settings.debug:
                    raise

            msg = (
                f"An error ocurred. If this keeps happening, please open an issue: "
                f"{utils.pekora_repo() / 'issues/new'}"
            )

            print(
                Panel(
                    msg,
                    title="Error",
                    title_align="left",
                    border_style="red",
                )
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pekora']

package_data = \
{'': ['*']}

install_requires = \
['alianator>=4.0.2,<5.0.0',
 'colour>=0.1.5,<0.2.0',
 'decorator>=5.1.1,<6.0.0',
 'inflect>=6.0.2,<7.0.0',
 'inquirerpy>=0.3.4,<0.4.0',
 'py-cord>=2.4.0,<3.0.0',
 'pyperclip==1.8.1',
 'typer[all]>=0.7.0,<0.8.0',
 'yarl>=1.8.2,<2.0.0']

entry_points = \
{'console_scripts': ['pekora = pekora.cli:app']}

setup_kwargs = {
    'name': 'pekora',
    'version': '1.0.4',
    'description': 'A command-line calculator for Discord permission values',
    'long_description': "# Pekora\n\n[![PyPI](https://img.shields.io/pypi/v/pekora?logo=pypi&color=green&logoColor=white&style=for-the-badge)](https://pypi.org/project/pekora)\n[![PyPI - License](https://img.shields.io/pypi/l/pekora?color=03cb98&style=for-the-badge)](https://github.com/celsiusnarhwal/pekora/blob/main/LICENSE.md)\n[![Code style: Black](https://aegis.celsiusnarhwal.dev/badge/black?style=for-the-badge)](https://github.com/psf/black)\n\nPekora is a calculator for Discord permission values. With it, you can calculate permission values, see detailed\ninformation about the permissions a value represents, and interactively build your own\npermissions.\n\n## Installation\n\n### pipx (recommended)\n\n[Install pipx](https://pypa.github.io/pipx/installation/), then run:\n\n```bash\npipx install pekora\n```\n\n### Homebrew (macOS and Linux only)\n\n[Homebrew](https://brew.sh) users can install Pekora from the\n[Houkago Tea Tap](https://github.com/celsiusnarhwal/homebrew-htt).\n\n```bash\nbrew tap celsiusnarhwal/htt\nbrew install pekora\n```\n\n## Documentation\n\nFor documentation, including usage instructions, visit [Pekora's website](https://pekora.celsiusnarhwal.dev).\n\n## License\n\nPekora is licensed under the [MIT License](https://github.com/celsiusnarhwal/pekora/blob/main/LICENSE.md). Its\ndocumentation is licensed under [CC BY 4.0](https://pekora.celsiusnarhwal.dev/license).\n",
    'author': 'celsius narhwal',
    'author_email': 'hello@celsiusnarhwal.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pekora.celsiusnarhwal.dev',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)
